"""Test functions from annotate."""
import contextlib
import pysam
import pytest
import os
from pore_c_py import annotate, utils


HEADER = pysam.AlignmentHeader.from_references(
    ["chr1", "chr2"], [1000, 2000])


def align_from_tuple(t, name="test", molecule_tag='unknown', walk_tag=None):
    """Align from tuple."""
    rec = pysam.AlignedSegment(header=HEADER)
    rec.reference_id = ["chr1", "chr2"].index(t[0])
    rec.reference_start = t[1]
    rec.query_sequence = "A" * (t[2] - t[1])
    rec.cigartuples = [(pysam.CMATCH, (t[2] - t[1]))]
    rec.flag = 0
    if t[3] == "-":
        rec.flag |= pysam.FREVERSE
    if t[3] == ".":
        rec.flag |= pysam.FUNMAP
    rec = utils.MonomerData(molecule_tag, t[4],  t[5], 0,0,0).to_pysam(rec)
    rec.query_name = name
    rec.set_tag(utils.WALK_TAG, walk_tag)
    return rec


def get_data_set(t, name="test", molecule_tag=None, walk_tag=None):
    """Get data set with different categories."""
    primary = align_from_tuple(t, name=name, molecule_tag=molecule_tag)
    unmapped = align_from_tuple(t, name=name, molecule_tag=molecule_tag)
    unmapped.is_unmapped = True
    secondary = align_from_tuple(t, name=name, molecule_tag=molecule_tag)
    secondary.is_secondary = True
    supplementary = align_from_tuple(t, name=name, molecule_tag=molecule_tag)
    supplementary.is_supplementary = True
    alns = [primary, secondary, supplementary, unmapped]  
    return (alns)


@pytest.mark.parametrize(
    "monomer,expected",
    [
        (("chr1", 0, 10, "+", 0, 10), "chr1:+:0-10:0-10"),
        (("chr2", 0, 20, "-", 0, 10), "chr2:-:0-20:0-10"),
        (("chr1", 0, 10, ".", 0, 10), "*:0-10"),
    ],
)
def test_get_walk_component(monomer, expected):
    """Test get monomer tag."""

    res = utils.MonomerData.from_pysam(align_from_tuple(monomer)).name
    assert res == expected


@pytest.mark.parametrize(
    "t",
    [
        (("chr1", 0, 10, "+", 0, 10))
    ],
)
def test_sort_by_category(t):
    """Test sort by category."""
    alns = get_data_set(t)
    primary = alns[0]
    res = annotate.sort_by_category(alns, "test")
    assert next(res) == primary


@pytest.mark.parametrize(
    "alns, expected_monomers, expected_tag",
    [
        ([("chr1", 0, 20, "+", 0, 10), ("chr1", 0, 5, "-", 0, 10)], 2,
         'chr1:+:0-20:0-10;chr1:-:0-5:0-10'),
        ([("chr1", 0, 10, "+", 0, 10), ("chr1", 9, 20, "+", 0, 20),
          ("chr1", 0, 10, ".", 0, 10)], 3, 
          "chr1:+:0-10:0-10;chr1:+:9-20:0-20;*:0-10")
    ]
)
def test_get_walk(alns, expected_monomers, expected_tag):
    """Test get walk."""
    aligns = []
    expected_walk = []
    for i, t in enumerate(alns):
        data_set = get_data_set(t, "test"+str(i), "test"+str(i))
        expected_walk += [data_set[0]]
        aligns += data_set
    walk, n_monomers = annotate.get_walk(aligns)
    assert walk == expected_walk
    assert n_monomers == expected_monomers
    assert walk[0].get_tag(utils.WALK_TAG) == expected_tag


@pytest.mark.parametrize(
    "alns, tag",
    [(
        [("chr1", 0, 20, "+", 0, 10),
         ("chr1", 0, 5, "-", 0, 10)],
        'chr1:+:0-20:0-10;chr1:-:0-5:0-10'),
     (
        [("chr1", 0, 20, "+", 0, 10),
         ("chr1", 0, 0, ".", 0, 10),
         ("chr1", 0, 10, "+", 0, 10)],
        'chr1:+:0-20:0-10;*:0-10;chr1:+:0-10:0-10')]
)
def test_annotate_alignments(alns, tag, tmpdir):
    """Test annotate alignments."""
    aligns = []
    expected_walk = []
    for i, t in enumerate(alns):
        data_set = get_data_set(t, "test"+str(i), "test")
        expected_walk += [align_from_tuple(t, "test"+str(i), "test", tag)]
        aligns += data_set
    bam_path = os.path.join(str(tmpdir), 'ex1.bam')
    with pysam.AlignmentFile(bam_path, mode='wb0', header=HEADER) as f:
        for i in aligns:
            f.write(i)
    with pysam.AlignmentFile(bam_path, "r", check_sq=False) as inbam:
        walks = list(annotate.annotate_alignments(inbam))
        assert len(walks) == 1
        assert walks[0] == expected_walk

#@pytest.mark.parametrize(
#    "start,end,read_length,expected",
#    [
#        (0, 1, 9, "concat001:0:1"),
#        (1, 9, 10, "concat001:01:09"),
#        (1, 10, 10, "concat001:01:10"),
#        # (0, 100, 1000,  "concat001:000"),
#        # (99, 100,10, "concat001:099"),
#    ],
#)
#def test_monomer_id(start, end, read_length, expected):
#    """Test monomer id."""
#    coords = ConcatemerCoords(
#        start=start, end=end,
#        read_length=read_length, subread_idx=0, subread_total=1
#    )
#    res = MonomerReadSeq.generate_id("concat001", coords)
#    assert res == expected
#
#
#def test_group_monomer_aligns(monomer_read_seqs):
#    """Test group monomer aligns."""
#    num_concatemers, aligns_per_concatemer = 2, 10
#    concatemer_ids = set()
#    for concat_id, monomers in group_aligns_by_concatemers(monomer_read_seqs):
#        assert len(list(monomers)) == aligns_per_concatemer
#        concatemer_ids.add(concat_id)
#    assert len(concatemer_ids) == num_concatemers
#    shuffled = sorted(
#        monomer_read_seqs, key=lambda x: x.read_seq.align_info.ref_name)
#    with pytest.raises(ValueError):
#        for concat_id, monomer_read_seqs in group_aligns_by_concatemers(shuffled): # noqa
#            pass
#
#
#def test_annotate_monomer_aligns(monomer_read_seqs):
#    """Test annotate monomer aligns."""
#    res = {
#        concat_id: aligns
#        for (concat_id, aligns) in annotate_monomer_alignments(monomer_read_seqs)   # noqa
#    }
#    assert len(res) == 2
#    for aligns in res.values():
#        assert len(aligns) == 10
#        for a in aligns:
#            a._update_tags()
#            tags = a.read_seq.tags
#            assert MOLECULE_TAG in tags
#            assert WALK_TAG in tags
#            assert CONCATEMER_TAG in tags
#
#
#def test_mods(mock_reads):
#    """Test mods."""
#    test_read = mock_reads["read_w_tags"]
#    header = AlignmentHeader.from_dict(
#        {"RG": [{"ID": "RG1", "SM": "sample01", "LB": "lib01"}]}
#    )
#    seg = test_read.to_align(header)
#    assert seg.modified_bases[("C", 0, "m")] == [(2, 122), (10, 128)]
#    # assert seg.get_tag("RG") == "RG1"
#
#
#def test_split_read(concatemer_unaligned: AlignedSegment):
#    """Test split read."""
#    concatemer = ConcatemerReadSeq.from_align(
#        concatemer_unaligned, as_unaligned=True, init_mod_bases=True
#    )
#    monomers = concatemer.split([5])
#    for idx, s in [(0, slice(None, 5)), (1, slice(5, None))]:
#        assert (
#            monomers[idx].read_seq.sequence
#            == concatemer_unaligned.seq[s]  # type: ignore
#        )
#        assert (
#            monomers[idx].read_seq.quality
#            == concatemer_unaligned.qual[s]  # type: ignore
#        )
#    assert monomers[0].to_align().modified_bases == {  # type: ignore
#        ("C", 0, "m"): [(2, 122)]
#    }
#    assert monomers[1].to_align().modified_bases == {  # type: ignore
#        ("C", 0, "m"): [(5, 128)]
#    }


"""Collection of file writers."""
import collections
import json
from typing import List

import pyarrow as pa
import pyarrow.parquet as pq
import pysam

from pore_c_py import align_tools, utils


class StatsWriter:
    """Stats writer."""

    def __init__(self, path):
        """Init."""
        self.path = path
        self.concatemer_count = 0
        self.cardinality_count = collections.Counter()
        self.pair_count = collections.Counter()
        self.cis_trans = collections.Counter()

    def append(self, pair):
        """Write records."""
        self.concatemer_count += 1
        cardinality = utils.MonomerData.from_pysam(pair.left).subread_idx
        if cardinality in self.cardinality_count.keys():
            self.cardinality_count[cardinality] += 1
        else:
            self.cardinality_count[cardinality] = 1
        self.pair_count[pair.state.name] += 1
        if pair.state == align_tools.PairState.both:
            if pair.left.reference_name == pair.right.reference_name:
                self.cis_trans["cis"] += 1
            else:
                self.cis_trans["trans"] += 1

    def close(self):
        """Close."""
        d = {
            "cardinality": self.cardinality_count,
            "pair_count": self.pair_count,
            "cis_trans": self.cis_trans
        }
        with self.path.open("w") as fh:
            fh.write(json.dumps(d))

    def __enter__(self):
        """Enter context."""
        pass

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit context, closing json file."""
        self.close()


class ChromunityWriter:
    """Chromunity writer."""

    def __init__(self, path, merge_distance=None):
        """Its the initializer init."""
        # TODO: see if chromunity can just parse this from the BAM
        self.path = path
        self.merge_distance = merge_distance
        self.schema = pa.schema([
            ("cid", pa.string()),
            ("read_name", pa.string()),
            ("chrom", pa.string()),
            ("start", pa.uint32()),
            ("end", pa.uint32()),
            ("num_fragments", pa.uint32()),
            ("pass_filter", pa.string())])
        self._writer = pq.ParquetWriter(str(self.path), self.schema)
        self.counter = 0

    def _aln_to_record(self, aln):
        return {
            "cid": aln.get_tag("MI"),
            "read_name": aln.get_tag("MI"),
            "chrom": aln.reference_name,
            "start": aln.reference_start,
            "end": aln.reference_end,
            "num_fragments": 1,
            "pass_filter": "TRUE"}

    def get_pylist(self, aligns):
        """Get pylist."""
        pylist = list()
        for block in align_tools.group_colinear(
                [r for r in aligns],
                tol=self.merge_distance):
            if len(block) == 1:
                if not block[0].is_unmapped:
                    pylist.append(self._aln_to_record(block[0]))
            else:
                rec = self._aln_to_record(block[0])
                rec["start"] = min(
                    block[0].reference_start, block[-1].reference_start)
                rec["end"] = max(
                    block[0].reference_end, block[-1].reference_end)
                rec["num_fragments"] = len(block)
                pylist.append(rec)
        return pylist

    def write(self, alignments: List[pysam.AlignedSegment]):
        """Write records."""
        pylist = list()
        if self.merge_distance is None:
            pylist = [
                self._aln_to_record(aln) for aln in alignments
                if not aln.is_unmapped]
        else:
            pylist = self.get_pylist(alignments)
        if len(pylist) > 0:
            batch = pa.RecordBatch.from_pylist(pylist, schema=self.schema)
            self._writer.write_batch(batch)
            self.counter += len(pylist)

    def close(self):
        """Close the parquet writer."""
        if self._writer is not None:
            self._writer.close()

    def __enter__(self):
        """Enter context."""
        # parquet writer is opened on creation
        pass

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exit context, closing parquet file."""
        self.close()



"""Command-line interface."""
import argparse
import contextlib
import itertools
import logging
from pathlib import Path
import sys
import time

import pysam

from pore_c_py import align_tools, annotate, digest, utils, writers


def porec_parser():
    """Create CLI parser."""
    from pore_c_py import __version__
    parser = argparse.ArgumentParser(
        "pore-c-py", parents=[utils.log_level()],
        description="Tools for processing Pore-C data.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True

    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {}'.format(__version__))

    digest_parse = subparsers.add_parser(
        'digest',
        description=(
            "Digest concatemer sequences into monomers "
            "using a restriction enzyme."),
        help=(
            "Resulting BAM file has query names of the format "
            "<original_read_id>:<subread_start>:<subread_end> with "
            "coordinates left padded with zeroes such that they are "
            f"sortable as strings. The f{utils.CONCATEMER_ID_TAG} SAM tag "
            "is set to the original "
            f"read/concatemer ID. The f{utils.MONOMER_DATA_TAG} tag "
            "contains the relative "
            "coordinates of the monomer on the concatemer. "
            f"{utils.MONOMER_DATA_TAG}:B:i,start,end,concatemer_length,subread_idx,subread_total "  # noqa:E501
            "If modified base tags are found in the input then they will "
            "processed so that each subread has the correct ML/MM tags. "
            "Note that the 'mv' SAM tag is not handled in the same way "
            "and is removed."),
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    digest_parse.set_defaults(func=digest_bam)
    digest_parse.add_argument(
        "input", type=Path, default=[sys.stdin], nargs='*',
        help="An unaligned BAM file or FASTQ file or a directory of same")
    digest_parse.add_argument(
        "enzyme",
        help="A restriction enzyme name eg. NlaIII")
    digest_parse.add_argument(
        "--output", type=Path, default=sys.stdout,
        help="An unaligned BAM file with a separate record for each monomer,"
        "default will write to standard out.")
    digest_parse.add_argument(
        "--header", type=Path,
        help="The output bam requires a header."
        "If INPUT parameter is stdin any existing headers will not be"
        "accessible so optionally provide an existing bam that"
        "contains headers to be copied to the output bam."
        "Otherwise the header will be 'pore-c-py'.")
    digest_parse.add_argument(
        "--recursive", default=False, action="store_true",
        help=(
            "If INPUT is a directory search recursively "
            "for files matching 'glob'")),
    digest_parse.add_argument(
        "--glob", default="*.bam",
        help="If INPUT is a directory use this glob to search for files")
    digest_parse.add_argument(
        "--remove_tags", nargs="+", default=set(),
        help="Optionally remove SAM tags from input")
    digest_parse.add_argument(
        "--max_monomers", type=int, default=None,
        help="Maximum number of monomers for a read to be included in output.")
    digest_parse.add_argument(
        "--excluded_list", type=Path, default=None,
        help="File output to write list of excluded reads.")
    digest_parse.add_argument(
        "--excluded_bam", type=Path, default=None,
        help="BAM output to write excluded reads.")
    digest_parse.add_argument(
        "--threads", default=1, type=int,
        help="Compute threads of bam compression.")

    annotate_parse = subparsers.add_parser(
        "annotate",
        description=(
            "Filter a BAM of concatemer-grouped monomer alignments, "
            "and add 'walk' tag."),
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    annotate_parse.set_defaults(func=annotate_bam)
    annotate_parse.add_argument(
        "bam", type=Path,
        help=(
            "A (concatemer) namesorted BAM of aligned "
            "and unaligned monomers."))
    annotate_parse.add_argument(
        "output_prefix", type=Path,
        help="Output files will share this prefix")
    annotate_parse.add_argument(
        "--stdout", default=False, action="store_true",
        help="Output to stdout")
    annotate_parse.add_argument(
        "--threads", default=1, type=int,
        help="Compute threads of bam compression.")
    annotate_parse.add_argument(
        "--force", default=False, action="store_true",
        help="Overwite any existing files")
    annotate_parse.add_argument(
        "--monomers", default=False, action="store_true",
        help="Create a namesorted BAM file with Pore-C annotations")
    annotate_parse.add_argument(
        "--chromunity", default=False, action="store_true",
        help="Create a chromunity-compatible parquet")
    annotate_parse.add_argument(
        "--chromunity_merge_distance", type=int,
        help=(
            "Merge co-linear monomers that are separated by less than this"
            "distance into a single monomer"))
    annotate_parse.add_argument(
        "--summary", default=False, action="store_true")
    annotate_parse.add_argument(
        "--direct_only", default=False, action="store_true",
        help=(
            "Only output monomer pairs that are adjacent on the concatemer"
            ", don't do combinatorial expansion."))
    annotate_parse.add_argument(
        "--paired_end", default=False, action="store_true",
        help="Create a mock paired-end BAM")
    annotate_parse.add_argument(
        "--filter_pairs", action="store_true",
        help=(
            "Filter out pairs using min and max distance"))
    annotate_parse.add_argument(
        "--paired_end_minimum_distance", default=0, type=int,
        help=(
            "Filter out any pairs shorter than this distance."
            "Note setting this removes all trans pairs."))
    annotate_parse.add_argument(
        "--paired_end_maximum_distance", default=float("inf"), type=int,
        help=(
            "Filter out any pairs longer than this distance. "
            "Note setting this removes all trans pairs."))
    annotate_parse.add_argument(
        "--allow_singletons", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow singletons in paired end bam"))
    annotate_parse.add_argument(
        "--allow_improper", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow improper pairs in paired end bam"))
    annotate_parse.add_argument(
        "--allow_unmapped", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow unmapped in paired end bam"))

    chunk_bam_parse = subparsers.add_parser(
        "chunk-bam", help="Chunk a BAM into pieces.",
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    chunk_bam_parse.set_defaults(func=chunk_bam)
    chunk_bam_parse.add_argument(
        "input", type=Path,
        help="An unaligned BAM file or directory of same")
    chunk_bam_parse.add_argument(
        "output_prefix", type=Path,
        help="An unaligned BAM file with a separate record for each monomer")
    chunk_bam_parse.add_argument(
        "--chunk_size", type=int, default=8000,
        help="The number of reads in each chunk"),
    chunk_bam_parse.add_argument(
        "--recursive", default=False, action="store_true",
        help=(
            "If INPUT is a directory search recursively for "
            "files matching 'glob'"))
    chunk_bam_parse.add_argument(
        "--glob", default="*.fastq",
        help="If INPUT is a directory use this glob to search for files")
    chunk_bam_parse.add_argument(
        "--max_reads", type=int,
        help="Take the first n reads (useful for testing)")
    chunk_bam_parse.add_argument(
        "--threads", default=1, type=int,
        help="Compute threads of bam compression.")

    return parser


def digest_bam(args):
    """Digest entry point.

    Read in concatemer sequences from unaligned bam or fastq.
    Cuts each read with enzyme provided,
    and tags with relevant concatemer info.
    Outputs a bam containing each monomer sequence.
    """
    logger = utils.get_named_logger("Digest")
    logger.info(f"Digesting concatemers from {args.input}.")
    # Default header will be pore-c-py.
    # If input type is a file or directory, Header will be copied from that.
    # If input type is stdin, there is an optional header parameter available
    # to supply a preferred header present in an existing bam.
    get_header = args.header
    input_files = [args.input[0]]
    input_mode = "rb"
    pysam_kwargs = {"text": "pore-c-py"}
    if sys.stdin not in args.input:
        input_files = list(
            utils.find_files(
                args.input[0], glob=args.glob, recursive=args.recursive))
        get_header = str(input_files[0])
        input_mode = "r"
    if args.header or sys.stdin not in args.input:
        with pysam.AlignmentFile(
                get_header, input_mode, check_sq=False) as inputfile:
            header = align_tools.update_header(inputfile.header)
        pysam_kwargs = {"header": header}
    mode = "wb"
    if (args.output is not sys.stdout) and \
            (not utils.stdout_is_regular_file()):
        mode = "wb0"
    with contextlib.ExitStack() as manager:
        outbam = pysam.AlignmentFile(
                args.output,
                threads=args.threads, mode=mode, **pysam_kwargs)
        manager.enter_context(outbam)
        excl_list = None
        if args.excluded_list is not None:
            excl_list = open(args.excluded_list, 'w')
            manager.enter_context(excl_list)
        excl_bam = None
        if args.excluded_bam is not None:
            excl_bam = pysam.AlignmentFile(
                args.excluded_bam,
                threads=args.threads, mode='w', **pysam_kwargs)
            manager.enter_context(excl_bam)

        for input_file in input_files:
            with pysam.AlignmentFile(
                    input_file, input_mode, check_sq=False) as inputfile:
                for is_monomer, read in digest.get_concatemer_seqs(
                    inputfile, enzyme=args.enzyme,
                        remove_tags=args.remove_tags,
                        max_monomers=args.max_monomers):
                    if is_monomer:
                        outbam.write(read)
                    else:
                        if excl_list is not None:
                            excl_list.write(f"{read.query_name}\n")
                        if excl_bam is not None:
                            excl_bam.write(read)
    logger.info("Finished digestion.")


def annotate_bam(args):
    """Parse BAM entry point."""
    logger = utils.get_named_logger("AnntateBAM")
    logger.info(f"Processing reads from {args.bam}")
    if all(
        x is False for x in (
            args.monomers, args.chromunity, args.summary, args.paired_end)):
        logger.error(
            "You must specify at least one of --monomers, --chromunity, "
            "--summary, or --paired_end")
        sys.exit(1)
    paired_end_bam = Path(f"{args.output_prefix}.pe.bam")
    summary_json = Path(f"{args.output_prefix}.summary.json")
    chromunity_parquet = Path(f"{args.output_prefix}.chromunity.parquet")
    outputs = [paired_end_bam, summary_json, chromunity_parquet]
    mode = "wb"
    if args.stdout:
        namesorted_bam = sys.stdout
        if not utils.stdout_is_regular_file():
            mode = "wb0"
    else:
        namesorted_bam = Path(f"{args.output_prefix}.ns.bam")
        outputs += namesorted_bam
    if any(x.exists() for x in outputs) and not args.force:
        logger.error(
            "Some of the outputs exist, please remove before continuing.")
        sys.exit(1)

    # TODO: other outputs
    with contextlib.ExitStack() as manager:
        inbam = pysam.AlignmentFile(args.bam, "r", check_sq=False)
        manager.enter_context(inbam)

        if args.monomers:
            header = align_tools.update_header(inbam.header)
            outbam = pysam.AlignmentFile(
                namesorted_bam, mode=mode, header=header, threads=args.threads)
            manager.enter_context(outbam)
        if args.chromunity:
            chrom_writer = writers.ChromunityWriter(
                chromunity_parquet,
                merge_distance=args.chromunity_merge_distance)
            manager.enter_context(chrom_writer)
        if args.paired_end:
            pairs_bam = pysam.AlignmentFile(
                paired_end_bam, "wb", header=header, threads=args.threads)
            manager.enter_context(pairs_bam)
        if args.summary:
            summary = writers.StatsWriter(summary_json)
            manager.enter_context(summary)

        for concat_walk in annotate.annotate_alignments(inbam):
            if args.monomers:
                for aln in concat_walk:
                    outbam.write(aln)
            if args.chromunity:
                chrom_writer.write(concat_walk)
            if args.summary or args.paired_end:
                pairs = align_tools.get_pairs(concat_walk, args.direct_only)
                if args.filter_pairs:
                    pairs = align_tools.filter_pairs(
                        pairs,
                        allow_singletons=args.allow_singletons,
                        allow_improper=args.allow_improper,
                        allow_unmapped=args.allow_unmapped)
                for p in pairs:
                    if args.paired_end:
                        pairs_bam.write(p.left)
                        if p.right is not None:
                            pairs_bam.write(p.right)
                    if args.summary:
                        summary.append(p)
    logger.info("Finished BAM parsing.")


def chunk_bam(args):
    """Create chunked ubam."""
    logger = utils.get_named_logger("ChunkedBAM")
    input_files = list(utils.find_files(
        args.input, glob=args.glob, recursive=args.recursive))
    if len(input_files) == 0:
        logger.error("No input files found at {args.input}")
        sys.exit(1)

    if args.output_prefix.is_dir():
        args.output_prefix = args.output_prefix / "chunked"

    with pysam.AlignmentFile(
            str(input_files[0]), "r", check_sq=False) as inbam:
        header = align_tools.update_header(inbam.header)

    read_stream = itertools.chain(
        *(pysam.AlignmentFile(str(x), check_sq=False) for x in input_files))
    if args.max_reads is not None:
        read_stream = itertools.islice(read_stream, args.max_reads)

    def _new_file(index, bam):
        if bam is not None:
            bam.close()
        return index + 1, pysam.AlignmentFile(
            str(args.output_prefix.with_suffix(f".batch_{index}.bam")),
            threads=args.threads, mode="wb", header=header)

    def _log_time(start, index):
        elapsed = time.perf_counter() - start
        reads_per_minute = int(args.chunk_size * 60 / elapsed)
        logger.info(
            f"Wrote batch {index}: {reads_per_minute} reads/min.")
        return time.perf_counter()

    bam = None
    index, written = 0, 0
    start = time.perf_counter()
    for align in read_stream:
        if written == 0:
            index, bam = _new_file(index, bam)
        bam.write(align)
        written += 1
        if written % args.chunk_size == 0:
            written = 0
            start = _log_time(start, index)
    if written != 0:
        _log_time(start, index)
    if bam is not None:
        bam.close()
    logger.info("Finished repacking BAMs.")


def run_main():
    """Entry point."""
    parser = porec_parser()
    args = parser.parse_args()

    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO)
    logger = logging.getLogger(__package__)
    logger.setLevel(args.log_level)
    if args.logfile:
        fh = logging.FileHandler(args.logfile, "w")
        fh.setLevel(args.log_level)
        logger.addHandler(fh)

    args.func(args)



"""Test functions in utils."""
from pathlib import Path

import pytest

from pore_c_py import utils


@pytest.mark.parametrize(
    "layout,glob,recursive,expected",
    [
        (["A/a/a.fastq"], None, None, None),
        (["a.fastq"], None, None, None),
        (["a.fastq.gz"], None, None, []),
        (["a.fastq", "A/a/a.fastq"], None, False, ["a.fastq"]),
    ],
)
def test_find_files(tmp_path, layout, glob, recursive, expected):
    """Test find files."""
    root = Path(str(tmp_path))
    if expected is None:
        expected = layout
    kwds = {}
    if glob is not None:
        kwds["glob"] = glob
    if recursive is not None:
        kwds["recursive"] = recursive
    for _ in layout:
        p = root / _
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(_)
    res = list([
        str(x).replace(str(tmp_path), "")[1:]
        for x in utils.find_files(root, **kwds)]
    )
    assert res == expected



"""Digestion of unaligned concatemers."""
import array
import copy

from Bio import Restriction
from Bio.Seq import Seq

from pore_c_py import utils


logger = utils.get_named_logger("Digest")


def get_subread_modified_bases(align, start, end):
    """Get modified bases subread.

    :param align: pysam.AlignedSegment.
    :param start: start coordinate to trim to.
    :param end: exclusive end co-ordinate.
    """
    # we use and array here since thats what pysam uses for the tag,
    # so might be a bit faster than an intermediate list.
    mm_str, ml_arr = "", array.array('B')
    base_indices = {}
    seq = align.query_sequence[start:end]
    for mod_key, mod_data in align.modified_bases.items():
        # find the modifications that overlap the subread
        idx = [
            x for x in range(len(mod_data)) if
            start <= mod_data[x][0] < end]
        probs_dic = dict(mod_data)
        if not idx:  # no mod bases (of this type) in the subread
            continue
        try:
            canonical_base, strand, skip_scheme, mod_type = mod_key
        except ValueError:
            canonical_base, strand, mod_type = mod_key
            skip_scheme = ""
        if canonical_base == "N":
            base_indices[canonical_base] = list(range(len(seq)))
        elif canonical_base not in base_indices:
            base_indices[canonical_base] = [
                x for x, b in enumerate(seq) if b.upper() == canonical_base
            ]
        base_offsets, probs = zip(*[mod_data[i] for i in idx])
        strand = "+" if strand == 0 else "-"
        deltas = []
        counter = 0
        for seq_idx in base_indices[canonical_base]:
            orig_idx = seq_idx + start
            if orig_idx in base_offsets:  # is modified
                ml_arr.append(probs_dic[orig_idx])
                deltas.append(str(counter))
                counter = 0
            else:
                counter += 1
            deltas_formatted = ','.join(deltas)
        mm_str += (
                f"{canonical_base}{strand}{mod_type}{skip_scheme}"
                f",{deltas_formatted};"
            )
    return mm_str, ml_arr


def splits_to_intervals(positions, length):
    """Split to intervals."""
    if len(positions) == 0:
        return [(0, length)]
    prefix, suffix = [], []
    if positions[0] != 0:
        prefix = [0]
    if positions[-1] != length:
        suffix = [length]
    breaks = prefix + positions + suffix
    return [(start, end) for start, end in zip(breaks[:-1], breaks[1:])]


def get_enzyme(enzyme):
    """Get enzyme."""
    enz = getattr(Restriction, enzyme, None)
    if enz is None:
        raise ValueError(f"Enzyme not found: {enzyme}")
    if enz.cut_twice():
        raise NotImplementedError(
            f"Enzyme cuts twice, not currently supported: {enzyme}"
        )
    return enz


def digest_sequence(align, enzyme, remove_tags=None, max_monomers=None):
    """Digest sequence.

    :param align: a pysam alignment.
    :param enzyme: biopython enzyme instance.
    :param remove_tags: set of tags to remove from alignment records.
    :param max_monomers: Maximum number of monomers for a read to be included
       in output.
    """
    if max_monomers is None:
        max_monomers = float('inf')
    if remove_tags is None:
        remove_tags = set()
    else:
        remove_tags = set(remove_tags)

    # build a set of tags to always remove regardless of the input.
    # This is everything produce by minimap2 plus the move 'mv'
    # from basecaller which is massive and not handled in trimming
    hardlist = {
        'tp', 'cm', 's1', 's2',
        'NM', 'MD', 'AS', 'SA',
        'ms', 'nn', 'ts', 'cg',
        'cs', 'dv', 'de', 'rl',
        'mv'}
    remove_tags.update(hardlist)
    # delete all mod tags if any were subject to deletion
    keep_mods = True
    mod_tags = {'Mm', 'Ml', 'MM', 'ML', 'Mn', 'MN'}
    if remove_tags & mod_tags:
        keep_mods = False
        remove_tags.update(mod_tags)

    # now actually remove stuff!
    for tag in remove_tags:
        if align.has_tag(tag):  # set tag doesn't like invalid tag codes
            align.set_tag(tag, None)

    concatemer_id = align.query_name
    cut_points = [x - 1 for x in enzyme.search(Seq(align.query_sequence))]
    read_length = len(align.query_sequence)
    num_digits = len(str(read_length))
    intervals = splits_to_intervals(cut_points, read_length)
    num_intervals = len(intervals)
    if num_intervals > max_monomers:
        # Quick return. We do this here and not a higher level to avoid making
        # lots of monomers below, which then would be discarded later.
        logger.warning(
            f"Dropping read {concatemer_id}, has {num_intervals} monomers.")
        yield False, copy.copy(align)
        return

    for idx, (start, end) in enumerate(intervals):
        read = copy.copy(align)
        # trim the sequence and quality
        seq = align.query_sequence[start:end]
        qual = None
        if align.query_qualities:
            qual = align.query_qualities[start:end]
        read.query_sequence = seq
        read.query_qualities = qual
        # deal with mods, upgrading tag from interim to approved spec
        if keep_mods:
            mm, ml = get_subread_modified_bases(align, start, end)
            # if keep mods reset any potential mod tags
            # for monomers that have no modified bases
            for tag in mod_tags:
                if read.has_tag(tag):  # set tag doesn't like invalid tag codes
                    read.set_tag(tag, None)
            if len(ml) == 0:
                logger.info(
                    f"Read {concatemer_id} has no modified bases.")
            else:
                read.set_tag("MM", mm)
                read.set_tag("ML", ml)
                read.set_tag("MN", len(seq))
        # lexograpically sortable monomer ID
        read.query_name = \
            f"{concatemer_id}:{start:0{num_digits}d}:{end:0{num_digits}d}"
        read.set_tag(
            utils.MONOMER_DATA_TAG,
            [start, end, read_length, idx, num_intervals])
        utils.MonomerData.set_monomer_data(
                read, start, end, read_length, idx, num_intervals)
        read.set_tag(utils.CONCATEMER_ID_TAG, concatemer_id, "Z")
        yield True, read


def get_concatemer_seqs(
        input_file, enzyme, remove_tags=None, max_monomers=None):
    """Digest concatemers in to unaligned monomers.

    :param input_file: pysam.AlignmentFile input
    :param enzyme: Name of the digestion enzyme used in sample preperation
    :param remove_tags: Comma seperated list of additional
        tags to remove from file
    :param max_monomers: Maximum number of monomers for a read to be included
       in output.

    Concatemers are split with chosen digestion enzyme in to monomers.
    Monomers are tagged with unique monomer id and concatemer info.
    """
    logger.info(f"Digesting unaligned sequences from {input_file}")
    n_concatemers = 0
    n_monomers = 0
    n_excluded = 0
    enzyme = get_enzyme(enzyme)
    tags_remove = {"mv"}
    if remove_tags:
        tags_remove.update(set(remove_tags))
    remove_tags = tags_remove
    for align in input_file.fetch(until_eof=True):
        n_concatemers += 1
        reads = digest_sequence(
            align, enzyme, remove_tags=remove_tags, max_monomers=max_monomers)
        for monomer, read in reads:
            if monomer:
                n_monomers += 1
            else:
                n_excluded += 1
            yield monomer, read
    logger.info(
        f"Found {n_monomers} monomers in {n_concatemers} concatemers.")
    logger.info(
        f"Excluded {n_excluded} concatemers.")



"""Things for operating on alignments."""
import copy
import enum
from itertools import combinations
import sys
from typing import List, Sequence

import pysam

import pore_c_py


class PairState(enum.Enum):
    """Pair State."""

    both = 0
    left = 1
    right = 2
    neither = 3
    singleton = 4


class PairedSegments:
    """Paired Segment class."""

    __align_state__ = {
        (False, False): PairState.both,
        (False, True): PairState.left,
        (True, False): PairState.right,
        (True, True): PairState.neither}

    def __init__(
            self,
            left: pysam.AlignedSegment,
            right: pysam.AlignedSegment):
        """Init."""
        # don't set these as self.left, self.right, just to save
        # some typing below
        left = copy.copy(left)
        right = copy.copy(right)

        if left is None:
            raise ValueError("Left read in pair cannot be None.")
        if right is None:  # singleton
            self.left = left
            self.right = None
            self.state = PairState.singleton
            return

        self.state = self.__align_state__[
            left.is_unmapped, right.is_unmapped]

        left.is_paired = True
        left.is_read1 = True
        left.mate_is_reverse = right.mate_is_reverse
        left.mate_is_unmapped = right.is_unmapped

        right.is_paired = True
        right.is_read2 = True
        right.mate_is_reverse = left.mate_is_reverse
        right.mate_is_unmapped = left.is_unmapped

        if self.state == PairState.both:
            if left.reference_name == right.reference_name:
                # TODO: check this: the spec says:
                # "each segment is properly aligned according to the aligner"
                # that doesn't strictly necessitate same reference for both
                left.is_proper_pair = True
                right.is_proper_pair = True
                left.next_reference_name = "="
                right.next_reference_name = "="
            else:
                left.next_reference_name = right.reference_name
                right.next_reference_name = left.reference_name
            left.next_reference_start = right.reference_start
            right.next_reference_start = left.reference_start
            mx = max(
                left.reference_end, right.reference_end)
            mn = min(
                left.reference_start, right.reference_start)
            template_length = mx - mn
            left.template_length = template_length
            right.template_length = -1 * template_length
        if self.state == PairState.left:
            left.mate_is_unmapped = True
            right.next_reference_name = left.reference_name
            right.next_reference_start = left.reference_start
        if self.state == PairState.right:
            right.mate_is_unmapped = True
            left.next_reference_name = right.reference_name
            left.next_reference_start = right.reference_start
        if self.state == PairState.neither:
            left.mate_is_unmapped = True
            right.mate_is_unmapped = True

        self.left = left
        self.right = right


def get_pairs(sorted_aligns, direct_only, keep_unpaired=True):
    """Get pairs."""
    if len(sorted_aligns) == 0:
        return
    elif len(sorted_aligns) == 1 and keep_unpaired:
        yield PairedSegments(sorted_aligns[0], None)
    else:
        if direct_only:
            pairs = zip(sorted_aligns[:-1], sorted_aligns[1:])
        else:
            pairs = combinations(sorted_aligns, 2)
        for left, right in pairs:
            yield PairedSegments(left, right)


def filter_pairs(
        pairs: Sequence[PairedSegments],
        min_distance=0, max_distance=float("inf"),
        allow_singletons=False, allow_improper=False, allow_unmapped=False):
    """Filter alignment pairs."""
    # TODO: something is weird here: filtering on distance, but then
    # allowing unmapped and improper is a bit weird. Should we stop
    # the caller from doing this?
    for pair in pairs:
        if pair.state == PairState.singleton:
            if allow_singletons:
                yield pair
            else:
                continue
        elif pair.state is not PairState.both:
            if allow_unmapped:
                yield pair
            else:
                continue
        # state == PairState.both
        distance = genomic_distance(pair.left, pair.right)
        if min_distance <= distance <= max_distance:
            yield pair


def genomic_distance(
        align1: pysam.AlignedSegment, align2: pysam.AlignedSegment):
    """Calculate distance between two alignments."""
    if (
            align1.is_unmapped or align2.is_unmapped
            or (align1.reference_name != align2.reference_name)):
        return float("inf")

    delta = min(align1.reference_end, align2.reference_end) \
        - max(align1.reference_start, align2.reference_start)
    return max(0, -delta)


def is_colinear(
        align1: pysam.AlignedSegment,
        align2: pysam.AlignedSegment, tol=0):
    """Check if two alignments are co-linear (and close) in the genome."""
    if align1.is_reverse != align2.is_reverse:
        return False
    distance = genomic_distance(align1, align2)
    return distance <= tol


def group_colinear(
        aligns: List[pysam.AlignedSegment], tol=0):
    """Group alignments into co-linear blocks."""
    if len(aligns) == 0:
        return []  # an empty list of blocks
    elif len(aligns) == 1:
        return [aligns]  # one block with single alignment
    res = []
    block = []
    last = aligns[0]
    block = [last]
    for aln in aligns[1:]:
        if is_colinear(last, aln, tol=tol):
            block.append(aln)
        else:
            res.append(block)
            block = [aln]
        last = aln
    res.append(block)
    return res


def update_header(header):
    """Add PG tag to existing BAM header.

    :param bamfile: a pysam.AlignmentFile.header
    """
    header = header.to_dict()
    pg = header.pop("PG", [])
    name = __package__.replace("_", "-")
    pg_data = {
        "ID": f"{name}-{len(pg) + 1}",
        "PN": name,
        "VN": pore_c_py.__version__,
        "CL": " ".join(sys.argv)}
    if len(pg) > 0:
        if "ID" in pg[-1]:
            pg_data["PP"] = pg[-1]["ID"]
    pg.append(pg_data)
    header["PG"] = pg
    return header



"""
pore_c_py setup script.
"""
import os
import pkg_resources
import re
from setuptools import setup, find_packages
import sys

__pkg_name__ = 'pore-c-py'
__author__ = 'epi2melabs'
__description__ = 'Analyse Pore-C data.'

__pkg_import__ = __pkg_name__.replace("-", "_")

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'

__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_import__))

# Get the version number from __init__.py
verstrline = open(os.path.join(__pkg_import__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_import__))

dir_path = os.path.dirname(__file__)
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    install_requires = [
        str(requirement) for requirement in
        pkg_resources.parse_requirements(fh)]

data_files = []
extra_requires = {}
extensions = []


setup(
    name=__pkg_name__,
    version=__version__,
    url='https://github.com/epi2me-labs/{}'.format(__pkg_name__),
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    zip_safe=False,
    install_requires=install_requires,
    packages=find_packages(exclude=("tests",)),
    entry_points={
        "console_scripts": [
            f"{__pkg_name__} = {__pkg_import__}.main:run_main"]},
)



"""Test functions from digest."""
import array
import copy

import pysam
import pytest

from pore_c_py import digest
from array import array
from tests.test_align_tools import align_from_tuple


def align_with_sequence(t, query_sequence=None, name="test"):
    """Align from tuple with specific sequence."""
    rec = align_from_tuple(t)
    if query_sequence is not None:
        rec.query_sequence = query_sequence
    rec.query_name = name
    rec.qual = "&" * (t[2] - t[1])
    return rec

@pytest.mark.parametrize(
    "length,positions,expected",
    [
        (10, [5], [(0, 5), (5, 10)]),
        (10, [0, 5], [(0, 5), (5, 10)]),
        (10, [0, 5, 10], [(0, 5), (5, 10)]),
        (10, [], [(0, 10)]),
        (10, [0], [(0, 10)]),
        (10, [0, 10], [(0, 10)]),
    ],
)
def test_splits_to_intervals(length, positions, expected):
    """Test splits to intervals."""
    intervals = digest.splits_to_intervals(positions, length)
    assert intervals == expected


# https://github.com/pysam-developers/pysam/blob/b0f1eb7f71b8f5305b112d26da5f72ce79dc33b3/tests/pysam_data/MM-chebi.sam
example1 = (
    ("*", 0, 36, "+"),
    # . = m
    # , = h (76792)
    # ; = m and h
    # | = N
    #      .        | .|,.          .  ;
    "AGCACTCCAGAGTCGNACGCCATYCGCGCGCCACCA",
    "C+m,2,2,1,4,1;C+76792,6,7;N+n,15,2;",
    # .   .   .   .   .   ,   ,   |   |
    [102,128,153,179,204,161,187,212,169])

@pytest.mark.parametrize(
    "name, sam_data, start, end, mm_exp, ml_exp",
    [
        # the whole read
        ("whole", example1, 0, 36, "C+m,2,2,1,4,1;C+76792,6,7;N+n,15,2;",
         [102, 128, 153, 179, 204, 161, 187, 212, 169]),

        # trim two non-C bases.
        # Doesn't effect C mods, rejigs coords of N
        ("trim-0C", example1, 2, 36, "C+m,2,2,1,4,1;C+76792,6,7;N+n,13,2;",
         [102, 128, 153, 179, 204, 161, 187, 212, 169]),

        # trim first C
        # Need to skip one less C for C mod subtag and for N (since its any base)
        ("trim-1C", example1, 3, 36, "C+m,1,2,1,4,1;C+76792,5,7;N+n,12,2;",
         [102, 128, 153, 179, 204, 161, 187, 212, 169]),

        # trim first two C
        # We now skip 0 Cs before first mod.
        ("trim-2C", example1, 5, 36, "C+m,0,2,1,4,1;C+76792,4,7;N+n,10,2;",
         [102, 128, 153, 179, 204, 161, 187, 212, 169]),

        # trim first three C
        # We've trimmed a mod base so drop the first entry from ML. Also
        ("trim-3C", example1, 7, 36, "C+m,2,1,4,1;C+76792,3,7;N+n,8,2;",
         [128, 153, 179, 204, 161, 187, 212, 169]),

        # trim last C
        # Affects both m and h codes
        ("trim+1C", example1, 0, 34, "C+m,2,2,1,4;C+76792,6;N+n,15,2;",
         [102, 128, 153, 179, 161, 212, 169]),

        # trim past first N
        # Affects both m and h codes also
        ("trim-5C", example1, 17, 36, "C+m,0,1,4,1;C+76792,1,7;N+n,1;",
         [128,153,179,204,161,187,169]),

        # trim past first N and last C
        # Affects both m and h codes also
        ("trim-5C+1C", example1, 17, 34, "C+m,0,1,4;C+76792,1;N+n,1;",
         [128,153,179,161,169]),

    ],
)
def test_get_subread_modified_bases(
        name, sam_data, start, end, mm_exp, ml_exp):
    """Test get subread."""
    # note: the representation isn't unique and the code doesn't produce the same
    #       representation as the test data. So we parse both using pysam and compare

    # create original
    aln_tuple, qseq, mm, ml = sam_data
    aln = align_with_sequence(aln_tuple, query_sequence=qseq, name="*")
    aln.set_tag("Mm", mm)
    aln.set_tag("Ml", ml)

    # create trimmed
    read = copy.copy(aln)
    mm, ml = digest.get_subread_modified_bases(aln, start, end)
    read.set_tag("Mm", mm)
    read.set_tag("Ml", ml)

    # original with expected tags
    expected_aln = align_with_sequence(aln_tuple, query_sequence=qseq, name="*")
    expected_aln.set_tag("Mm", mm_exp)
    expected_aln.set_tag("Ml", ml_exp)

    # compare
    for k, v in read.modified_bases.items():
        assert expected_aln.modified_bases[k] == v

    # check that ML tag is a byte array
    assert isinstance(read.get_tag("Ml"), array)
    found = sum(x.startswith('Ml:B:C') for x in read.to_dict()["tags"])
    assert found == 1


@pytest.mark.parametrize(
    "align, sequence, enzyme, expected",
    [
        (("chr1", 0, 19, "+"), 'AAACATGTTTGGCATGAAA', "NlaIII",
         ['AAACATG', 'TTTGGCATG', 'AAA']),
        (("chr1", 0, 10, "+"), 'GGATCTGATC', "DpnII",
         ['G', 'GATCT', 'GATC'])
    ],
)
def test_digest_sequence(align, sequence, enzyme, expected):
    aln = align_with_sequence(align, query_sequence=sequence)
    enz = digest.get_enzyme(enzyme)
    seqs = digest.digest_sequence(aln, enz)
    for i, (monomer, gen) in enumerate(seqs):
        assert monomer
        assert gen.query_sequence == expected[i]


@pytest.mark.parametrize(
    "align, sequence, MM, ML, expected_MN, expected_ML, expected_MM",
    [(
         ("*", 0, 43, "+"),
         "AGCACTCCAGAGTCGNACGCCATYCGCGCGCCACCACATGTTT",
         "C+m,2,2,1,4,1;C+76792,6,7;N+n,15,2;",
         [102,128,153,179,204,161,187,212,169],
         [40, 0], 
         [[102, 128, 153, 179, 204, 212, 169, 161, 187]],
         ['C+m,2,2,1,4,1;N+n,15,2;C+76792,6,7;']),
     (
         ("*", 0, 43, "+"),
         "AGCACTCCAGAGTCGNACGCCATYCGCGCGCCACCACATGTTC",
         "C+m,2,2,1,4,1,1;C+76792,6,7;N+n,15,2;",
         [102,128,153,179,204,161,187,212,169,100],
         [40, 3],
         [[102, 128, 153, 179, 204, 169, 100, 187, 212],
          [161]],
         ['C+m,2,2,1,4,1;N+n,15,2;C+76792,6,7;', 'C+m,0;'])
    ],
)
def test_digest_sequence_mods(align, sequence, MM, ML, expected_MN, expected_ML, expected_MM):
    # Create aln
    aln = align_with_sequence(align, query_sequence=sequence)
    aln.set_tag("MN", len(sequence))
    aln.set_tag("Mn", len(sequence))
    aln.set_tag("Ml", ML)
    aln.set_tag("Mm", MM)

    # Get digested with modified bases
    enz = digest.get_enzyme("NlaIII")
    seqs = digest.digest_sequence(aln, enz)
 
    # Compare 
    for i, (monomer, gen) in enumerate(seqs):
        assert gen.has_tag("Mn") == False
        if gen.has_tag("MN"):
            assert gen.get_tag("MN") == expected_MN[i]
            assert gen.get_tag("ML") == array('B', expected_ML[i])
            assert gen.get_tag("MM") == expected_MM[i]
        else:
            assert gen.has_tag("MN") == False
        
        


@pytest.mark.parametrize(
    "align, sequence, enzyme",
    [
        (("chr1", 0, 19, "+"), 'AAACATGTTTGGCATGAAA', "NlaIII"),
    ],
)
def test_remove_mv_tag(align, sequence, enzyme):
    aln = align_with_sequence(align, query_sequence=sequence)
    aln.set_tag('mv', [0,1,0,1,0,1])
    enz = digest.get_enzyme(enzyme)
    seqs = digest.digest_sequence(aln, enz)
    for (monomer, gen) in seqs:
        assert not gen.has_tag('mv')

    # check also we don't blow up removing tag that doesn't exist
    seqs = digest.digest_sequence(aln, enz, remove_tags={'thing'})
    for (monomer, gen) in seqs:
        assert not gen.has_tag('mv')




"""Tests to align_tools."""
import pysam
import pytest

from pore_c_py import align_tools


HEADER = pysam.AlignmentHeader.from_references(
    ["chr1", "chr2", "*"], [1000, 2000, 3000])

def align_from_tuple(t):
    """Align from tuple."""
    rec = pysam.AlignedSegment(header=HEADER)
    rec.reference_id = ["chr1", "chr2", "*"].index(t[0])
    rec.reference_start = t[1]
    rec.query_sequence = "A" * (t[2] - t[1])
    rec.cigartuples = [(pysam.CMATCH, (t[2] - t[1]))]
    rec.flag = 0
    if t[3] == "-":
        rec.flag |= pysam.FREVERSE
    if t[3] == ".":
        rec.flag |= pysam.FUNMAP
    return rec


@pytest.mark.parametrize(
    "left,right,tol,expected",
    [
        # overlapping
        (("chr1", 0, 10, "+"), ("chr1", 0, 20, "+"), 0, True),
        (("chr1", 0, 10, "+"), ("chr1", 0, 20, "-"), 0, False),
        (("chr1", 0, 10, "-"), ("chr1", 0, 20, "-"), 0, True),
        (("chr1", 0, 10, "+"), ("chr2", 0, 20, "+"), 0, False),
        # need tol=1 for bookended features
        (("chr1", 0, 10, "+"), ("chr1", 10, 20, "+"), 1, True),
        # allow some wiggle
        (("chr1", 0, 10, "+"), ("chr1", 5, 20, "+"), 5, True),
        # unaligned not colinear
        (("chr1", 0, 10, "+"), ("chr1", 0, 0, "."), 0, False)
    ],
)
def test_is_colinear(left, right, tol, expected):
    """Test is colinear."""
    res = align_tools.is_colinear(
        align_from_tuple(left),
        align_from_tuple(right),
        tol=tol,)
    assert res == expected


@pytest.mark.parametrize(
    "aligns,expected",
    [
        [
            [("chr1", 0, 10, "+"), ("chr1", 9, 20, "+"), ("chr2", 0, 10, "+")],
            [[0, 1], [2]],
        ],
        [
            [("chr1", 0, 8, "+"), ("chr1", 8, 9, "."), ("chr1", 20, 30, "+")],
            [[0], [1], [2]],
        ],
        [
            [("chr1", 0, 8, "+"), ("chr1", 8, 9, ".")],
            [[0], [1]],
        ],
        [
            [("chr1", 0, 8, "+")],
            [[0]],
        ],
        [
            [],
            []
        ]
    ],
)
def test_group_colinear(aligns, expected):
    """Test group colinear."""
    alns = [align_from_tuple(t) for t in aligns]
    exp = list()
    for grp in expected:
        exp.append([alns[x] for x in grp])
    res = align_tools.group_colinear(alns)
    assert res == exp


@pytest.mark.parametrize(
    "left,right,expected",
    [   # overlapping
        (("chr1", 0, 5, "+"), ("chr1", 2, 15, "+"), 0),
        # not overlapping
        (("chr1", 0, 5, "+"), ("chr1", 10, 15, "+"), 5),
        (("chr1", 10, 15, "-"), ("chr1", 0, 5, "+"), 5),
        (("chr1", 0, 5, "-"), ("chr1", 10, 15, "+"), 5),
        (("chr1", 10, 15, "-"), ("chr1", 0, 5, "+"), 5),
        (("chr1", 0, 10, "+"), ("chr1", 11, 15, "-"), 1),
        (("chr1", 0, 5, "+"), ("chr1", 10, 15, "-"), 5),
        (("chr1", 10, 15, "+"), ("chr1", 0, 5, "-"), 5),
        (("chr1", 0, 5, "-"), ("chr1", 10, 15, "-"), 5),
        (("chr1", 10, 15, "+"), ("chr1", 0, 5, "+"), 5),
        # unmapped
        (("chr1", 0, 10, "+"), ("chr1", 11, 15, "."), float("inf")),
        (("chr1", 0, 10, "."), ("chr1", 11, 15, "."), float("inf")),

    ],
)
def test_genomic_distance(left, right, expected):
    """Test genomic distance."""

    res = align_tools.genomic_distance(
        align_from_tuple(left), align_from_tuple(right))
    assert res == expected


@pytest.mark.parametrize(
    "left, right, expected_left_flags, expected_right_flags",
    [
        # Neither
        (("chr1", 0, 5, "."), ("chr1", 0, 5, "."),
         {"mate_is_unmapped": True}, {"mate_is_unmapped": True}),
        # Both same ref
        (("chr1", 0, 5, "+"), ("chr1", 10, 20, "-"),
         {"is_proper_pair": True, "next_reference_name": "chr1",
         "next_reference_start": 10, "template_length": 20},
         {"is_proper_pair": True, "next_reference_name": "chr1",
         'next_reference_start': 0, 'template_length': -20}),
        # Both different ref
        (("chr1", 0, 5, "+"), ("chr2", 10, 20, "+"),
         {"is_proper_pair": False, "next_reference_name": "chr2",
          "next_reference_start": 10},
         {"is_proper_pair": False, "next_reference_name": "chr1",
          "next_reference_start": 0}),
        # right
        (("chr1", 0, 5, "."), ("chr2", 10, 20, "+"),
         {"mate_is_unmapped": False, "next_reference_name": "chr2",
          "next_reference_start": 10},
         {"mate_is_unmapped": True, "next_reference_name": None,
          "next_reference_start": -1}),
        # left
        (("chr1", 0, 5, "+"),("chr2", 10, 20, "."),
         {"mate_is_unmapped": True, "next_reference_name": None,
         "next_reference_start": -1},
         {"mate_is_unmapped": False, "next_reference_name": "chr1",
         "next_reference_start": 0}),
        # singleton
        (("chr1", 0, 5, "+"), None, {}, {})
    ],
)
def test_paired_segments(
        left, right, expected_left_flags, expected_right_flags):
    """Test paired segments class."""
    if right is not None:
        right = align_from_tuple(right)
    left = align_from_tuple(left)
    paired_segments = align_tools.PairedSegments(left, right)
    for k, v in expected_left_flags.items():
        assert getattr(paired_segments.left, k) == v
    for k, v in expected_right_flags.items():
        assert getattr(paired_segments.right, k) == v
    if right is None:
        assert paired_segments.state == align_tools.PairState.singleton
        assert paired_segments.right is None


@pytest.mark.parametrize(
    "aligns,  direct, expected, expected_segments",
    [   
        ([("chr1", 0, 10, "+"), ("chr1", 9, 20, "+"), ("chr2", 0, 10, "+")], False,
         align_tools.PairState.both, 3),
        ([("chr1", 0, 10, "+"), ("chr1", 0, 10, "+"), ("chr1", 9, 20, "+"),
          ("chr1", 9, 20, "+"), ("chr2", 0, 10, "+")], False,
         align_tools.PairState.both, 10),
        ([("chr1", 0, 10, "+"), ("chr1", 0, 10, "+"), ("chr1", 9, 20, "+"),
          ("chr1", 9, 20, "+"), ("chr2", 0, 10, "+")], True,
        align_tools.PairState.both, 4),
        ([], False, None, None),
        ([("chr2", 0, 10, "+")], False, align_tools.PairState.singleton, 1),
        ],
)
def test_get_pairs(aligns,  direct, expected, expected_segments):
    """Test get pairs."""
    aligns = [align_from_tuple(t) for t in aligns]
    pair_segments = align_tools.get_pairs(aligns, direct_only=direct)
    if len(aligns) != 0:
        count = 0
        for i in pair_segments:
            count += 1
            assert i.state == expected
        assert count == expected_segments
    else:
        assert next(pair_segments, None) == expected


@pytest.mark.parametrize(
    "expected",
    [
        ("pore-c-py")
    ],
)
def test_update_header(expected):
    """Test update header."""
    header = align_tools.update_header(HEADER)
    assert header['PG'][-1]['PN'] == expected


# def test_short_walk(m1, caplog):
#     """Test short walk."""
#     m1[2].read_seq.flags.secondary = True
#     with caplog.at_level(logging.WARNING):
#         m = list(
#             annotate_monomer_alignments(m1[:-1], remove_qcfail=False))[0][1]
#         qcfail = [_.read_seq.flags.qcfail for _ in m]
#         assert qcfail == [False, False, True]
#         assert (
#             caplog.record_tuples[-1][2]
#             == "Expected to see 3 alignments for concatemer C, found 2"
#         )
#
#
# @pytest.fixture
# def m2():
#     """Make mock monomer aligns m2."""
#     return (
#         _make_mock_monomer_aligns(
#             [
#                 (0, 5, "chr1", 0, None),
#                 (5, 10, "chr2", 100, None),
#                 (20, 25, "chr1", 50, None),
#                 (25, 30, "chr1", 500, None),
#             ]
#         ),
#         [
#             PairData(*_)
#             for _ in [
#                 ("C", 0, True, 0, PairAlignState.both, False, None, None),
#                 ("C", 1, True, 10, PairAlignState.both, False, None, None),
#                 ("C", 2, True, 0, PairAlignState.both, True, 445, True),
#             ]
#         ],
#     )
#
#
# def test_pairs_direct(m2):
#     """Test pairs direct."""
#     aligns, expected = m2
#     _, _, pair_data = zip(*get_pairs(aligns, direct_only=True))
#     assert list(pair_data) == expected
#
#
# @pytest.fixture
# def m3():
#     """Make mock monomer aligns m3."""
#     return (
#         _make_mock_monomer_aligns(
#             [
#                 (0, 5, "chr1", 0, None),
#                 (5, 10, "chr2", 100, None),
#                 (20, 25, "chr1", 50, None),
#             ]
#         ),
#         [
#             PairData(*_)
#             for _ in [
#                 ("C", 0, True, 0, PairAlignState.both, False, None, None),
#                 ("C", 1, False, 15, PairAlignState.both, True, 45, True),
#                 ("C", 2, True, 10, PairAlignState.both, False, None, None),
#             ]
#         ],
#     )
#
#
# def test_pairs_indirect(m3):
#     """Test pairs indirect."""
#     aligns, expected = m3
#     _, _, pair_data = zip(*get_pairs(aligns, direct_only=False))
#     assert list(pair_data) == expected



"""Annotation of monomer alignments."""
from itertools import groupby

from pore_c_py import utils

logger = utils.get_named_logger("AnntateAln")


def sort_by_category(alns, monomer_id):
    """Sort by category."""
    reordered = list()
    for aln in alns:
        # original claims this as "primary", thats not technically accurate
        category = 0
        if aln.is_secondary:
            category = 3
        elif aln.is_supplementary:
            category = 2
        elif aln.is_unmapped:
            category = 1
        reordered.append(list([aln, category]))
    # create a generator to be consistent with input
    reordered.sort(key=lambda a: a[1])
    if reordered[0][1] not in {0, 1}:
        logger.warning(
            f"Warning: best alignment for monomer: {monomer_id} "
            f"is secondary or supplementary")
    yield from (x[0] for x in reordered)


def get_walk(aligns):
    """Get walk."""
    n_monomers = 0
    walk = list()
    for monomer_id, alns in groupby(aligns, lambda x: x.query_name):
        n_monomers += 1
        sorted_aligns = sort_by_category(alns, monomer_id)
        aln = next(sorted_aligns)
        walk.append(aln)
    expected_monomers = utils.MonomerData.get_subread_total(walk[-1])
    if expected_monomers != len(walk):
        concat_id = walk[-1].get_tag(utils.CONCATEMER_ID_TAG)
        logger.warning(
            f"Expected to see {expected_monomers} alignments for "
            f"concatemer {concat_id}, found {len(walk)}"
        )
    walk_tag = ";".join(utils.MonomerData.from_pysam(aln).name for aln in walk)
    for aln in walk:
        aln.set_tag(utils.WALK_TAG, walk_tag)
    return (walk, n_monomers)


def annotate_alignments(bamfile):
    """Annotate (and filter) monomer alignments.

    :param bamfile: pysam.AlignmentFile
    :yields: the file header following by lists of alignments by concatemer.

    Alignments are grouped by concatemer (the input should be sorted by
    concatemer stored in the concatemer ID tag), and one alignment is selected
    per monomer. Alignments are annotated with a "walk", which simply
    enumerates the alignment coordinates of the monomers comprising the
    concatemer.
    """
    total_monomers = 0
    seen = set()
    for concat_id, aligns in groupby(
            bamfile.fetch(until_eof=True),
            lambda x: x.get_tag(utils.CONCATEMER_ID_TAG)):
        if concat_id in seen:
            logger.warning(
                f"Input file appears unsorted, found {concat_id} "
                "more than once.")
        seen.add(concat_id)
        walk, n_monomers = get_walk(aligns)
        total_monomers += n_monomers
        yield walk
    logger.info(f"Found {total_monomers} monomers in {len(seen)} concatemers.")



"""Log."""
import argparse
from dataclasses import dataclass
import logging
import os
from pathlib import Path
import stat
import sys

import pysam


def log_level():
    """Parser to set logging level and acquire software version/commit."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    modify_log_level = parser.add_mutually_exclusive_group()
    modify_log_level.add_argument(
        '--debug', action='store_const',
        dest='log_level', const=logging.DEBUG, default=logging.INFO,
        help='Verbose logging of debug information.')
    modify_log_level.add_argument(
        '--quiet', action='store_const',
        dest='log_level', const=logging.WARNING, default=logging.INFO,
        help='Minimal logging; warnings only.')
    parser.add_argument("--logfile", type=Path, help="Specify a log file.")

    return parser


def get_named_logger(name):
    """Create a logger with a name.

    :param name: name of logger.
    """
    name = name.ljust(10)[:10]  # so logging is aligned
    logger = logging.getLogger('{}.{}'.format(__package__, name))
    logger.name = name
    return logger


def find_files(root: Path, glob="*.fastq", recursive=True):
    """Find files."""
    if not root.is_dir():
        yield root
    else:
        if recursive and not glob.startswith("**/"):
            glob = f"**/{glob}"
        for f in root.glob(glob):
            yield (f)


def stdout_is_regular_file() -> bool:
    """
    Detect if standard output is a regular file (or say a pipe).

    :return: True if stdout is a regular file, else False.
    """
    mode = os.fstat(sys.stdout.buffer.fileno()).st_mode
    return stat.S_ISREG(mode)


CONCATEMER_ID_TAG = "MI"
MONOMER_DATA_TAG = "Xc"
WALK_TAG = "Xw"

# NOTE: the class below exists only because we don't store individual bits in
#       separate tags; which would be simpler and more direct to access.


@dataclass
class MonomerData:
    """Concatemer Data."""

    concatemer_id: str
    start: int
    end: int
    read_length: int
    subread_idx: int
    subread_total: int

    @classmethod
    def from_pysam(cls, align: pysam.AlignedSegment):
        """Extract concatemer meta information from alignment."""
        cls.concatemer_id = align.get_tag(CONCATEMER_ID_TAG)
        tags = align.get_tag(MONOMER_DATA_TAG)
        cls.start = tags[0]
        cls.end = tags[1]
        cls.read_length = tags[2]
        cls.subread_idx = tags[3]
        cls.subread_total = tags[4]

        if align.is_unmapped:
            cls.name = (
                f"*:"
                f"{cls.start}-{cls.end}")
        else:
            orientation = "+-"[align.is_reverse]
            cls.name = (
                f"{align.reference_name}:{orientation}:"
                f"{align.reference_start}-{align.reference_end}:"
                f"{cls.start}-{cls.end}")
        return cls

    def to_pysam(self, align: pysam.AlignedSegment):
        """Set concatemer metadata on alignment."""
        align.set_tag(CONCATEMER_ID_TAG, self.concatemer_id)
        align.set_tag(MONOMER_DATA_TAG, [
            self.start, self.end, self.read_length,
            self.subread_idx, self.subread_total])
        return align

    @staticmethod
    def set_monomer_data(
            align: pysam.AlignedSegment,
            start, end, read_length, idx, num_intervals):
        """Set the monomer data on an alignment."""
        align.set_tag(
            MONOMER_DATA_TAG,
            [start, end, read_length, idx, num_intervals])

    @staticmethod
    def get_subread_total(align: pysam.AlignedSegment):
        """Return the number of sibling monomers."""
        return align.get_tag(MONOMER_DATA_TAG)[4]



"""Test functions in writers."""
import pytest
import os
from pathlib import Path
from pore_c_py import align_tools, writers, utils
from tests.test_align_tools import align_from_tuple


def monomer_tagged_align(t, name="*", molecule_tag=None):
    """Align from tuple."""
    rec = align_from_tuple(t)
    rec = utils.MonomerData(
       molecule_tag, t[4],  t[5], 0, 0, 0).to_pysam(rec)
    rec.query_name = name
    rec.set_tag(utils.CONCATEMER_ID_TAG, molecule_tag)
    return rec


@pytest.mark.parametrize(
    "aligns, expected",
    [
        ([("chr1", 0, 10, "+", 10, 20), ("chr1", 9, 20, "+", 0, 11),
          ("chr2", 0, 10, "+", 5, 15 )],
         {'cardinality': 3, 'concatemer_count': 3,
          'pair_count': {'both': 3}, 'cis_trans':{'trans': 2,'cis': 1}}),
        ([("chr2", 0, 10, "+", 10, 20)],
         {'cardinality': 1, 'concatemer_count': 1,
          'pair_count':{'singleton':1}, 'cis_trans': {}}),
    ],
)
def test_stats_writer_append(aligns, expected):
    "Test stats writer append."
    aligns = [monomer_tagged_align(t, molecule_tag="test") for t in aligns]
    pair_segments = align_tools.get_pairs(aligns, direct_only=False)
    st = Path("tmp.json")
    stats = writers.StatsWriter(st)
    for i in pair_segments:
        stats.append(i)
    assert stats.cardinality_count[0] == expected['cardinality']
    assert stats.concatemer_count == expected['concatemer_count']
    assert stats.pair_count == expected['pair_count']
    assert stats.cis_trans == expected['cis_trans']


@pytest.mark.parametrize(
    "aligns, expected",
    [
        ([("chr1", 0, 10, "+", 10,20), ("chr1", 0, 20, "+", 0,20), ("chr1", 9, 20, "+", 0,11)], 3),
        ([("chr1", 0, 10, "+", 10,20), ("chr1", 0, 20, ".", 0,20)], 1),
        ([("chr1", 0, 10, "+", 10,20), ("chr1", 0, 20, "+", 0,20)], 2)
    ],
)
def test_chromunity_get_pylist(aligns, expected, tmpdir):
    "Test Chromunity get pylist."
    aligns = [monomer_tagged_align(t, molecule_tag="test") for t in aligns]
    chr_json = os.path.join(str(tmpdir), 'tmp.json')
    chrom = writers.ChromunityWriter(chr_json, merge_distance=5)
    test = chrom.get_pylist(aligns)
    assert test[0]['num_fragments'] == expected
    

@pytest.mark.parametrize(
    "aligns, expected",
    [
        ([("chr1", 0, 8, "+", 0, 8), ("chr1", 50, 60, "+",50, 60),
          ("chr1", 20, 30, "+",20, 30)], 3),
        ([("chr1", 0, 10, "+", 0, 10), ("chr2", 0, 10, "+",8 , 9)], 2),
    ],
)
def test_chromunity_writer(aligns, expected, tmpdir):
    "Test Chromunity writer."
    aligns = [monomer_tagged_align(t, molecule_tag="test") for t in aligns]
    chr_json = os.path.join(str(tmpdir), 'tmp.json')
    chrom = writers.ChromunityWriter(chr_json, merge_distance=5)
    chrom.write(aligns)
    assert chrom.counter == expected